import os
import json
import subprocess
import time
import pytest
import boto3
from datetime import datetime
from urllib.parse import urlparse

assert "AWS_DEFAULT_REGION" in os.environ
assert os.environ.get("MINIWDL__AWS__FSAP", "").startswith(
    "fsap-"
), "set environment MINIWDL__AWS__FSAP to EFS Access Point ID (fsap-xxxx)"
assert (
    "MINIWDL__AWS__WORKFLOW_IMAGE" in os.environ
    and "miniwdl-aws" in os.environ["MINIWDL__AWS__WORKFLOW_IMAGE"]
), "set environment MINIWDL__AWS__WORKFLOW_IMAGE to repo:digest"
assert (
    "MINIWDL__AWS__WORKFLOW_QUEUE" in os.environ
), "set MINIWDL__AWS__WORKFLOW_QUEUE to Batch queue name"
assert "MINIWDL__AWS__TASK_QUEUE" in os.environ, "set MINIWDL__AWS__TASK_QUEUE to Batch queue name"
assert (
    "MINIWDL_AWS_TEST_BUCKET" in os.environ
), "set MINIWDL_AWS_TEST_BUCKET to test S3 bucket (name only)"


@pytest.fixture(scope="module")
def aws_batch():
    return boto3.client("batch", region_name=os.environ["AWS_DEFAULT_REGION"])


def batch_miniwdl(aws_batch, args, environment=None, upload=None, cache=False):
    """
    Submit & await a Batch job to run cmd in the miniwdl_aws container (usually ~miniwdl run~
    to launch other Batch jobs in turn)
    """
    cmd = ["python3", "-m", "miniwdl_aws"]
    cmd.extend(args)
    cmd.append("--follow")
    if not cache:
        cmd.append("--no-cache")
    if upload:
        if not upload.endswith("/"):
            upload += "/"
        cmd.extend(["--s3upload", upload])

    exit_code = subprocess.run(
        cmd, cwd=os.path.dirname(os.path.dirname(__file__)), check=False
    ).returncode

    if exit_code != 0:
        ans = {"success": False, "exit_code": exit_code}
        if upload:
            error = get_s3uri(upload + "error.json")
            if error:
                ans["error"] = json.loads(error)
        return ans

    ans = {"success": True}
    if upload:
        outputs = get_s3uri(upload + "outputs.json")
        if outputs:
            ans["outputs"] = json.loads(outputs)
    return ans


def get_s3uri(uri):
    """
    Download bytes from s3:// URI
    """
    try:
        assert uri.startswith("s3://")
        parts = urlparse(uri)
        obj = boto3.resource("s3", region_name=os.environ["AWS_DEFAULT_REGION"]).Object(
            parts.netloc, parts.path.lstrip("/")
        )
        return obj.get()["Body"].read()
    except Exception as exn:
        if "NoSuchKey" in str(exn):
            return None
        raise


def test_miniwdl_run_self_test(aws_batch):
    subprocess.run(
        ["python3", "-m", "miniwdl_aws", "--follow", "--self-test", "--no-cache"],
        cwd=os.path.dirname(os.path.dirname(__file__)),
        check=True,
    )


@pytest.fixture(scope="session")
def test_s3_folder():
    """
    S3 folder for this test session
    """
    return f"s3://{os.environ['MINIWDL_AWS_TEST_BUCKET']}/{datetime.today().strftime('%Y%m%d_%H%M%S')}/"


def test_retry_streams(aws_batch, test_s3_folder):
    rslt = batch_miniwdl(
        aws_batch,
        [
            "/var/miniwdl_aws_test_assets/test_retry_streams.wdl",
            "--dir",
            "/mnt/efs/miniwdl_aws_tests",
            "--verbose",
        ],
        upload=test_s3_folder + "test_retry_streams/",
    )
    assert rslt["success"]
    assert len(rslt["outputs"]["test_retry_streams.messages"]) == 4
    assert len(rslt["outputs"]["test_retry_streams.stdouts"]) == 4
    assert len(rslt["outputs"]["test_retry_streams.stderrs"]) == 4
    for i in range(4):
        assert (
            get_s3uri(rslt["outputs"]["test_retry_streams.messages"][i]).decode().strip()
            == "Hello, stdout!"
        )
        assert (
            get_s3uri(rslt["outputs"]["test_retry_streams.stdouts"][i]).decode().strip()
            == "Hello, stdout!"
        )
        assert (
            get_s3uri(rslt["outputs"]["test_retry_streams.stderrs"][i]).decode().strip()
            == "Hello, stderr!"
        )


def test_assemble_refbased(aws_batch, test_s3_folder):
    rslt = batch_miniwdl(
        aws_batch,
        [
            "https://github.com/broadinstitute/viral-pipelines/raw/v2.1.19.0/pipes/WDL/workflows/assemble_refbased.wdl",
            "reads_unmapped_bams=https://github.com/broadinstitute/viral-pipelines/raw/v2.1.19.0/test/input/G5012.3.testreads.bam",
            "reference_fasta=https://github.com/broadinstitute/viral-pipelines/raw/v2.1.19.0/test/input/ebov-makona.fasta",
            "sample_name=G5012.3",
            "--dir",
            "/mnt/efs/miniwdl_aws_tests",
            "--verbose",
        ],
        upload=test_s3_folder + "test_assemble_refbased/",
    )
    assert rslt["success"]
    # TODO: more assertions


def test_termination(aws_batch, test_s3_folder):
    """
    The workflow will fail and should do so relatively quickly (modulo worker instance startup time)
    """
    t0 = time.time()
    rslt = batch_miniwdl(
        aws_batch,
        [
            "/var/miniwdl_aws_test_assets/test_termination.wdl",
            "--dir",
            "/mnt/efs/miniwdl_aws_tests",
            "--verbose",
        ],
        upload=test_s3_folder + "test_termination/",
    )
    assert not rslt["success"]
    assert (
        "This is the end, my only friend"
        in get_s3uri(rslt["error"]["cause"]["stderr_s3file"]).decode()
    )
    assert time.time() - t0 < 600


def test_nonexistent_docker(aws_batch, test_s3_folder):
    """
    Workflow specifies a docker image that doesn't exist; does this error bubble up from AWS Batch
    in a reasonable way?
    """
    rslt = batch_miniwdl(
        aws_batch,
        [
            "/var/miniwdl_aws_test_assets/test_nonexistent_docker.wdl",
            "docker=nonexistent_bogus_12345",
            "--dir",
            "/mnt/efs/miniwdl_aws_tests",
            "--delete-after",
            "failure",
            "--verbose",
        ],
        upload=test_s3_folder + "test_nonexistent_docker/",
    )
    assert not rslt["success"]
    assert "CannotPullContainerError" in str(rslt["error"])


def test_call_cache(aws_batch, test_s3_folder):
    """
    Call cache works (short-term, where previous outputs remain on /mnt/shared)
    """
    t0 = int(time.time())
    # run once to prime cache
    rslt = batch_miniwdl(
        aws_batch,
        [
            "/var/miniwdl_aws_test_assets/test_call_cache.wdl",
            "timestamp_in=",
            str(t0),
            "names=Alice",
            "names=Bob",
            "names=Carol",
            "fail=true",
            "--verbose",
            "--dir",
            "/mnt/efs/miniwdl_aws_tests",
        ],
        cache=False,
    )
    assert not rslt["success"]

    # run again where a subset of calls should be reused
    t1 = int(time.time())
    rslt = batch_miniwdl(
        aws_batch,
        [
            "/var/miniwdl_aws_test_assets/test_call_cache.wdl",
            "timestamp_in=",
            str(t0),
            "names=Alice",
            "names=Bob",
            "names=Xavier",
            "--verbose",
            "--dir",
            "/mnt/efs/miniwdl_aws_tests",
        ],
        cache=True,
        upload=test_s3_folder + "test_call_cache/",
    )
    assert rslt["success"]

    # Alice and Bob were cached, Xavier was not:
    assert t0 <= rslt["outputs"]["test_call_cache.timestamps_out"][0] <= t1
    assert t0 <= rslt["outputs"]["test_call_cache.timestamps_out"][1] <= t1
    assert rslt["outputs"]["test_call_cache.timestamps_out"][2] > t1
    assert "Hello, Alice!" in get_s3uri(rslt["outputs"]["test_call_cache.messages"][0]).decode()
    assert "Hello, Bob!" in get_s3uri(rslt["outputs"]["test_call_cache.messages"][1]).decode()
    assert "Hello, Xavier!" in get_s3uri(rslt["outputs"]["test_call_cache.messages"][2]).decode()


def test_call_cache_one_task(aws_batch, test_s3_folder):
    """
    Short-term call cache of one task (where the entire run outputs, not just a portion thereof,
    are sourced from the cache.)
    """
    t0 = int(time.time())
    rslt = batch_miniwdl(
        aws_batch,
        [
            "/var/miniwdl_aws_test_assets/test_call_cache.wdl",
            "timestamp_in=",
            str(t0),
            "name=Alyssa",
            "--task",
            "write_name",
            "--verbose",
            "--dir",
            "/mnt/efs/miniwdl_aws_tests",
        ],
        cache=False,
    )
    assert rslt["success"]

    t1 = int(time.time())
    rslt = batch_miniwdl(
        aws_batch,
        [
            "/var/miniwdl_aws_test_assets/test_call_cache.wdl",
            "timestamp_in=",
            str(t0),
            "name=Alyssa",
            "--task",
            "write_name",
            "--verbose",
            "--dir",
            "/mnt/efs/miniwdl_aws_tests",
        ],
        cache=True,
        upload=test_s3_folder + "test_call_cache_one_task/",
    )
    assert rslt["success"]

    assert t0 <= rslt["outputs"]["write_name.timestamp_out"] <= t1
    assert "Alyssa" in get_s3uri(rslt["outputs"]["write_name.name_file"]).decode()


def test_download(aws_batch):
    """
    Test workflow can use https:// and s3:// input files. This is functionality built-in to miniwdl
    so ought to just work, but nice to cover it here.
    """
    rslt = batch_miniwdl(
        aws_batch,
        [
            "/var/miniwdl_aws_test_assets/count_lines.wdl",
            "files=https://raw.githubusercontent.com/chanzuckerberg/miniwdl/main/tests/alyssa_ben.txt",
            "files=s3://1000genomes/CHANGELOG",
            "--dir",
            "/mnt/efs/miniwdl_aws_tests",
            "--verbose",
        ],
    )
    assert rslt["success"]



import os
import boto3
import base64
import json
import uuid
import requests
from WDL._util import StructuredLogMessage as _


def detect_aws_region(cfg):
    if cfg and cfg.has_option("aws", "region") and cfg.get("aws", "region"):
        return cfg.get("aws", "region")

    # check environment variables
    for ev in ("AWS_REGION", "AWS_DEFAULT_REGION"):
        if os.environ.get(ev):
            return os.environ[ev]

    # check boto3, which will load ~/.aws
    if boto3.DEFAULT_SESSION and boto3.DEFAULT_SESSION.region_name:
        return boto3.DEFAULT_SESSION.region_name
    session = boto3.Session()
    if session.region_name:
        return session.region_name

    # query EC2 metadata
    try:
        return requests.get(
            "http://169.254.169.254/latest/meta-data/placement/region", timeout=2.0
        ).text
    except:
        pass

    return None


def randomize_job_name(job_name):
    # Append entropy to the Batch job name to avoid race condition using identical names in
    # concurrent RegisterJobDefinition requests
    return (
        job_name[:103]  # 119 + 1 + 8 = 128
        + "-"
        + base64.b32encode(uuid.uuid4().bytes[:5]).lower().decode()
    )


def efs_id_from_access_point(region_name, fsap_id):
    # Resolve the EFS access point id (fsap-xxxx) to the associated file system id (fs-xxxx). Saves
    # user from having to specify both.
    aws_efs = boto3.Session().client("efs", region_name=region_name)
    desc = aws_efs.describe_access_points(AccessPointId=fsap_id)
    assert len(desc.get("AccessPoints", [])) == 1
    desc = desc["AccessPoints"][0]
    fs_id = desc["FileSystemId"]
    assert isinstance(fs_id, str) and fs_id.startswith("fs-")
    return fs_id


def detect_sagemaker_studio_efs(logger, **kwargs):
    # Detect if we're operating inside SageMaker Studio and if so, record EFS mount details
    METADATA_FILE = "/opt/ml/metadata/resource-metadata.json"
    metadata = None
    try:
        with open(METADATA_FILE) as infile:
            metadata = json.load(infile)
        assert metadata["DomainId"] and metadata["UserProfileName"]
    except:
        return None
    try:
        api = boto3.client("sagemaker", **kwargs)
        domain = api.describe_domain(DomainId=metadata["DomainId"])
        efs_id = domain["HomeEfsFileSystemId"]
        profile = api.describe_user_profile(
            DomainId=metadata["DomainId"], UserProfileName=metadata["UserProfileName"]
        )
        efs_uid = profile["HomeEfsFileSystemUid"]
        assert efs_id and efs_uid
        efs_home = f"/{efs_uid}/miniwdl"
        efs_mount = os.path.join(os.getenv("HOME"), "miniwdl")
        logger.notice(
            _(
                "detected SageMaker Studio",
                domain=metadata["DomainId"],
                user=metadata["UserProfileName"],
                efs_id=efs_id,
                efs_home=efs_home,
                efs_mount=efs_mount,
            )
        )
        return (efs_id, efs_uid, efs_home, efs_mount)
    except Exception as exn:
        logger.warning(
            _(
                "detected local AWS SageMaker Studio metadata, but failed to query domain",
                error=str(exn),
                metadata=METADATA_FILE,
                domain=metadata["DomainId"],
                user=metadata["UserProfileName"],
            )
        )
        return None


def detect_studio_fsap(logger, efs_id, efs_uid, efs_home, **kwargs):
    # Look for an Access Point with the appropriate configuration to mount the SageMaker Studio EFS
    # (in the same way it's presented through Studio)
    try:
        efs = boto3.client("efs", **kwargs)
        access_points = efs.describe_access_points(FileSystemId=efs_id, MaxResults=100).get(
            "AccessPoints", []
        )
        if len(access_points) >= 100:
            logger.warn(
                _(
                    "EFS has >=100 Access Points; set configuration [aws] fsap or environment MINIWDL__AWS__FSAP to avoid searching through them",
                    efs_id=efs_id,
                )
            )
        for ap in access_points:
            assert ap["FileSystemId"] == efs_id
            if (
                ap["LifeCycleState"] == "available"
                and ap.get("RootDirectory", {}).get("Path", "") == efs_home
                and str(ap.get("PosixUser", {}).get("Uid", "")) == efs_uid
            ):
                logger.notice(
                    _(
                        "detected suitable EFS Access Point; to override, set configuration [aws] fsap or environment MINIWDL__AWS__FSAP",
                        arn=ap["AccessPointArn"],
                    )
                )
                return ap["AccessPointId"]
        return None
    except Exception as exn:
        logger.warning(
            _(
                "error detecting EFS Access Point",
                error=str(exn),
                efs_id=efs_id,
                ufs_uid=efs_uid,
            )
        )
        return None


def detect_gwfcore_batch_queue(logger, efs_id, **kwargs):
    # Look for a Batch job queue tagged with the Studio EFS id (indicating it's our default)
    try:
        batch = boto3.client("batch", **kwargs)
        queues = batch.describe_job_queues(maxResults=100).get("jobQueues", [])
        if len(queues) >= 100:
            logger.warn(
                "AWS Batch has >=100 job queues; set configuration [aws] task_queue or environment MINIWDL__AWS__TASK_QUEUE to avoid searching through them"
            )
        queues = [
            q
            for q in queues
            if q.get("state", "") == "ENABLED"
            and q.get("status", "") == "VALID"
            and q.get("tags", {}).get("MiniwdlStudioEfsId", "") == efs_id
        ]
        if not queues:
            return None
        if len(queues) > 1:
            default_queues = [q for q in queues if q.get("jobQueueName", "").startswith("default-")]
            if default_queues:
                queues = default_queues
        logger.notice(
            _(
                "detected suitable AWS Batch job queue; to override, set configuration [aws] task_queue or environment MINIWDL__AWS__TASK_QUEUE",
                arn=queues[0]["jobQueueArn"],
            )
        )
        return queues[0]["jobQueueName"]
    except Exception as exn:
        logger.warning(
            _(
                "error detecting AWS Batch job queue",
                error=str(exn),
                efs_id=efs_id,
            )
        )
        return None


END_OF_LOG = "[miniwdl_run_s3upload] -- END OF LOG --"



from setuptools import setup, find_packages
from version import get_version

with open("README.md") as fp:
    long_description = fp.read()

setup(
    name="miniwdl-aws",
    version=get_version(),
    description="miniwdl AWS backend (Batch+EFS)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Wid L. Hacker",
    python_requires=">=3.6",
    packages=find_packages(),
    setup_requires=["reentry"],
    install_requires=["miniwdl>=1.2.1", "boto3>=1.17", "requests"],
    reentry_register=True,
    entry_points={
        "miniwdl.plugin.container_backend": ["aws_batch_job = miniwdl_aws:BatchJob"],
        "console_scripts": [
            "miniwdl-run-s3upload = miniwdl_aws:miniwdl_run_s3upload",
            "miniwdl-aws-submit = miniwdl_aws.__main__:main",
        ],
    },
)
